@@ -1,12 +1,15 @@
 use crate::melpow::hash;
 
-use std::convert::TryInto;
-use std::fmt;
-
 use rustc_hash::FxHashMap;
 use smallvec::SmallVec;
+use std::collections::HashMap;
+use std::convert::TryInto;
+use std::fmt;
+
+use assoc::AssocExt;
+
+pub type SVec<T> = SmallVec<[T; 40]>;
 
-pub type SVec<T> = SmallVec<[T; 32]>;
 #[derive(Copy, Clone, Ord, PartialOrd, Eq, PartialEq, Hash)]
 pub struct Node {
     pub bv: u64,
@@ -40,22 +43,23 @@ impl Node {
         self.bv >> n & 1
     }
 
-    pub fn get_parents(self, n: usize) -> SVec<Node> {
-        let mut parents = SVec::new();
+    pub fn get_parents(&self, n: usize) -> Vec<Node> {
+        let mut v = vec![];
+        self.foreach_parent(n, |p| v.push(p));
+        v
+    }
+
+    pub fn foreach_parent(self, n: usize, mut f: impl FnMut(Node)) {
         if self.len == n {
-            let range = 0..n;
-
-            range.into_iter().for_each(|index| {
+            for index in 0..n {
                 if (self.bv >> index) & 1 != 0 {
-                    parents.push(self.take(index).append(0))
+                    f(self.take(index).append(0))
                 }
-            });
+            }
         } else {
-            parents.push(self.append(0));
-            parents.push(self.append(1));
+            f(self.append(0));
+            f(self.append(1));
         }
-
-        parents
     }
 
     pub fn uniqid(self) -> u64 {
@@ -108,8 +112,49 @@ impl fmt::Debug for Node {
 
 pub fn calc_labels(chi: &[u8], n: usize, f: &mut impl FnMut(Node, &[u8])) {
     calc_labels_helper(chi, n, Node::new_zero(), f, &mut FxHashMap::default());
+    // // iterative implementation
+    // let mut memoizer: FxHashMap<Node, SVec<u8>> = FxHashMap::default();
+    // // let mut memoizer: Vec<(Node, SVec<u8>)> = Default::default();
+    // let mut stack = Vec::with_capacity(32);
+    // stack.push((false, Node::new_zero()));
+    // while let Some((revisit, nd)) = stack.pop() {
+    //     // eprintln!(
+    //     //     "visiting {} at stack size {} and memoizer size {} ",
+    //     //     nd,
+    //     //     stack.len(),
+    //     //     memoizer.len()
+    //     // );
+    //     if nd.len == n {
+    //         let mut lab_gen = hash::Accumulator::new(chi);
+    //         lab_gen.add(&nd.to_bytes());
+    //         nd.foreach_parent(n, |parent| {
+    //             lab_gen.add(memoizer.get(&parent).unwrap());
+    //         });
+
+    //         let lab = lab_gen.hash();
+    //         f(nd, &lab);
+    //         memoizer.insert(nd, lab);
+    //     } else if !revisit {
+    //         stack.push((true, nd));
+    //         stack.push((false, nd.append(1)));
+    //         stack.push((false, nd.append(0)));
+    //     } else {
+    //         let l0 = memoizer.get(&nd.append(0)).unwrap().clone();
+    //         let l1 = memoizer.get(&nd.append(1)).unwrap().clone();
+    //         memoizer.remove(&nd.append(0));
+    //         memoizer.remove(&nd.append(1));
+    //         let lab = hash::Accumulator::new(chi)
+    //             .add(&nd.to_bytes())
+    //             .add(&l0)
+    //             .add(&l1)
+    //             .hash();
+    //         f(nd, &lab);
+    //         memoizer.insert(nd, lab);
+    //     }
+    // }
 }
 
+#[inline]
 fn calc_labels_helper(
     chi: &[u8],
     n: usize,
@@ -120,10 +165,8 @@ fn calc_labels_helper(
     if nd.len == n {
         let mut lab_gen = hash::Accumulator::new(chi);
         lab_gen.add(&nd.to_bytes());
-        let parents = nd.get_parents(n);
-
-        parents.iter().for_each(|parent| {
-            lab_gen.add(&ell[parent]);
+        nd.foreach_parent(n, |parent| {
+            lab_gen.add(&ell[&parent]);
         });
 
         let lab = lab_gen.hash();
