@@ -16,10 +16,11 @@ use std::{collections::HashSet, fmt::Debug};
 use std::{convert::TryInto, sync::Arc};
 
 use crate::state::melswap::PoolMapping;
-use dashmap::lock::RwLock;
+use dashmap::{lock::RwLock, DashMap};
 use defmac::defmac;
 use derivative::Derivative;
 use novasmt::{dense::DenseMerkleTree, ContentAddrStore};
+use once_cell::sync::Lazy;
 use stdcode::StdcodeSerializeExt;
 use tap::Pipe;
 use themelio_structs::{
@@ -288,7 +289,7 @@ impl<C: ContentAddrStore> State<C> {
                 .insert_coin(pseudocoin_id, pseudocoin_data, self.tip_906());
         }
         // create the finalized state
-        SealedState(self, action, Default::default())
+        SealedState(self, action)
     }
 }
 
@@ -296,16 +297,12 @@ impl<C: ContentAddrStore> State<C> {
 /// It cannot be constructed except through sealing a State or restoring from persistent storage.
 #[derive(Derivative, Debug)]
 #[derivative(Clone(bound = ""))]
-pub struct SealedState<C: ContentAddrStore>(
-    State<C>,
-    Option<ProposerAction>,
-    Arc<RwLock<Option<State<C>>>>,
-);
+pub struct SealedState<C: ContentAddrStore>(State<C>, Option<ProposerAction>);
 
 impl<C: ContentAddrStore> SealedState<C> {
     /// From raw parts
     pub fn from_parts(state: State<C>, prop_action: Option<ProposerAction>) -> Self {
-        Self(state, prop_action, Default::default())
+        Self(state, prop_action)
     }
 
     /// Returns a reference to the State finalized within.
@@ -327,11 +324,7 @@ impl<C: ContentAddrStore> SealedState<C> {
     /// Decodes from the partial encoding.
     pub fn from_partial_encoding_infallible(bts: &[u8], db: &novasmt::Database<C>) -> Self {
         let tmp: (Vec<u8>, Option<ProposerAction>) = stdcode::deserialize(bts).unwrap();
-        SealedState(
-            State::from_partial_encoding_infallible(&tmp.0, db),
-            tmp.1,
-            Default::default(),
-        )
+        SealedState(State::from_partial_encoding_infallible(&tmp.0, db), tmp.1)
     }
 
     /// Returns the block header represented by the finalized state.
@@ -380,36 +373,39 @@ impl<C: ContentAddrStore> SealedState<C> {
     }
     /// Creates a new unfinalized state representing the next block.
     pub fn next_state(&self) -> State<C> {
-        let next = self.2.read();
-        if let Some(next) = next.clone() {
-            return next;
-        }
-        drop(next);
-        let mut new = State::clone(self.inner_ref());
-        // fee variables
-        new.history.insert(self.0.height, self.header());
-        new.height += BlockHeight(1);
-        new.stakes.remove_stale((new.height / STAKE_EPOCH).0);
-        new.transactions.clear();
-        // TIP-906 transition
-        if new.tip_906() && !self.inner_ref().tip_906() {
-            log::warn!("DOING TIP-906 TRANSITION NOW!");
-            let old_tree = new.coins.inner().clone();
-            let mut count = old_tree.count();
-            for (_, v) in old_tree.iter() {
-                let cdh: CoinDataHeight =
-                    stdcode::deserialize(&v).expect("pre-tip906 coin tree has non-cdh elements?!");
-                let old_count = new.coins.coin_count(cdh.coin_data.covhash);
-                new.coins
-                    .insert_coin_count(cdh.coin_data.covhash, old_count + 1);
-                if count % 100 == 0 {
-                    log::warn!("{} left", count);
+        static CACHE: Lazy<DashMap<HashVal, Vec<u8>>> = Lazy::new(Default::default);
+        if let Some(v) = CACHE.get(&self.header().hash()) {
+            State::from_partial_encoding_infallible(
+                v.value(),
+                &self.inner_ref().coins.inner().database(),
+            )
+        } else {
+            let mut new = State::clone(self.inner_ref());
+            // fee variables
+            new.history.insert(self.0.height, self.header());
+            new.height += BlockHeight(1);
+            new.stakes.remove_stale((new.height / STAKE_EPOCH).0);
+            new.transactions.clear();
+            // TIP-906 transition
+            if new.tip_906() && !self.inner_ref().tip_906() {
+                log::warn!("DOING TIP-906 TRANSITION NOW!");
+                let old_tree = new.coins.inner().clone();
+                let mut count = old_tree.count();
+                for (_, v) in old_tree.iter() {
+                    let cdh: CoinDataHeight = stdcode::deserialize(&v)
+                        .expect("pre-tip906 coin tree has non-cdh elements?!");
+                    let old_count = new.coins.coin_count(cdh.coin_data.covhash);
+                    new.coins
+                        .insert_coin_count(cdh.coin_data.covhash, old_count + 1);
+                    if count % 100 == 0 {
+                        log::warn!("{} left", count);
+                    }
+                    count -= 1;
                 }
-                count -= 1;
+                CACHE.insert(self.header().hash(), new.partial_encoding());
             }
+            new
         }
-        *self.2.write() = Some(new.clone());
-        new
     }
 
     /// Applies a block to this state.
